#include <hpgcc49.h>
#include "minimax.h"



int error=0; // global variable for error reporting


int main()
{
	int retval;
	retval= do_minimax();		// minimax polynom
	sat_push_zint_llong(error);
	return 0;
}

int //  Valentin's  minimax  2009-09-24 
do_minimax(){
	int m,m2,n, n0;
	double **data, *pol, abserr, eps;
//	sat3_ckn(3);	// require 3 elements on calculator stack
//	sat3_checkargs(3, SAT_DOREAL, SAT_DOREAL, SAT_DOARRY);
	eps=sat_pop_real();
	n0=sat_pop_real(); if (n0<1) n0=1;
	m2=sat_pop_zint_llong();
	m=sat_pop_zint_llong();
    data=create_matrix2(m, m2);
	pop_matrix_reals(m, m2, data);
	if (m2!=2) {error=Invalid_Dimensions;return(DO_ERROR_EXIT); }		// "Error invalid dimensions"
	if (n0>m-1) {error=Bad_Argument_Value;return(DO_ERROR_EXIT); }		// "Error invalid dimensions"


	pol=create_vector((m+2));
    n=minimax(pol,data,&abserr,eps,m,n0);
	free_matrix2(m,2,data);
	free(pol);
	return (0);
}

int minimax(double *pol, double **data, double *c, double eps,int m, int n)
// * minimax polynom iteration 
// * Valentin's minimax 
//  pol : current minimax polynom, allocated size m+1, for worst case 
//	data: [x,y], dim {m,2} in Valentin's program designated z[m] and y[m]
//	c	: max absoulute error of (pol(x)-y)
//	eps	: allowable error 
//	m	: dimension of data
//	n	: initial size of pol  (increases if necessary)

{
//	const int w=1;
//	int retval; 
	int nsys, ll, i=0, j, t, n0=n; 
	int finish_flag=0;
	double err=0.0,h, k, h0, err0;  
	double *a=0, *u=0, *v=0, *d=0;
	const double b=0.001;
	d=create_vector(m);if (!d) return (DO_ERROR_EXIT);
	c=&err;
	n--;
	do{
		n++;
		h0=data[0][0];
		err0=data[0][1];
		nsys=n+2; 
		ll=nsys-1;
		if(a)free(a);a=create_vector(nsys);
		if(u)free(u);u=create_vector(nsys);
		if(v)free(v);v=create_vector(nsys);
		if (!a || !u || !v ) return (DO_ERROR_EXIT);
//		clear_screen();
		for(i=0;i<nsys; i++){ 
			j=(i*2*(m-1)+ll)/(2*ll); // to ensure same selection as rpl implementation
			u[i]=data[j][0];
			v[i]=data[j][1];
		} 
		finish_flag=0;
		while (!(finish_flag)){
			spolmm(u,v,a,nsys);
			h=fabs(a[0]);
			vec_fnerr(d, m, a, data, nsys);
			t=fmaxabs(d, m);
			err=d[t];
			k=data[t][0];
		 	if (n==m-1||err==0.0) finish_flag=1; 
		 	else finish_flag=( fabs(fabs(h/err)-1)<b || (h==h0 && err==err0));
			if(finish_flag) continue; // drop rest of loop, then exit loop
			h0=h;err0=err;
			if (u[0]<k && k <u[ll]){
			 	i=-1; 
			    while(u[++i]<=k);
			    if (fns(a, nsys, u, v, i)!=fsign(err)) i--;
			    u[i]=k;
			    v[i]=data[t][1];
			}
			else if (k>=u[ll]){ 
	    	   	if (fns(a, nsys, u, v,ll)!=fsign(err)){
					for (i=0;i<ll;i++){ u[i]=u[i+1]; v[i]=v[i+1];} 
   				}
			    u[ll]=k;
			    v[ll]=data[t][1];
			}
			else{ 
				if (fns(a, nsys, u, v, 0)!=fsign(err)){
					for (i=ll;i>0;i--){ u[i]=u[i-1]; v[i]=v[i-1];}
 				}
 				u[0]=k;
			    v[0]=data[t][1];
			}   
		}  // next tuple

	    for (i=0;i<=n;i++) pol[n-i]=a[i+1];
	    push_result_to_stack((n+1), (fabs(a[0])),pol);
	}while ( !(eps<0 || fabs(err)<eps ||n==m-1));    // do-while finished->final result
	sat_push_real((n-n0+1));	// total number of lists pushed to the stack
	free(d);
	free(u);
	free(v);
	free(a);
	return(n);	// degree of final polynom 
}   

int spolmm(double *u, double *v, double *a,int n)
{
	int i,j,retval=0, *ipvt,sgn=1;
	double z,zn, **work, **mat;
	ipvt=create_int_vector(n);  
	mat=create_matrix2(n,n);
	work=create_matrix2(n,1);
	if (!ipvt || !mat || !work) return(DO_ERROR_EXIT);
	for (i=0; i<n;i++){ 
		work[i][0]=v[i];
		sgn*=-1; 
		zn=1.0; 
		z=u[i];  // x[i]
		mat[i][0]=sgn;
		mat[i][1]=1.0;
		for (j=2;j<n;j++){
			zn*=z;
			mat[i][j]=zn;
			}
	}
	if (dgefa(n, mat, ipvt)) { error=(Infinite_Result); retval=DO_ERROR_EXIT; goto exit_spolmm;}		// "singular"
	dgesl(1, n, mat, ipvt, work, 0);
	for (i=0; i<n;i++)a[i]=work[i][0];
exit_spolmm:		
	free_matrix2(n,n, mat);		
	free_matrix2(n,1, work);
	free(ipvt);	
	return (retval);
}

double fnf(double *a, double x, int n)
{   int i;
    double r;
    r=0.0;for (i=n-1; i>=1; i--) r=r*x+a[i];
 return (r);    
}	

int fns(double *pol, int n, double *u, double *v, int i)
{   
   double r;
    r=fnf(pol,u[i],n);
    r=r-v[i];
    return (fsign(r));    
}	

int fsign( double x)
{
	int i=0;
	if (x>0) i=1;
	if (x<0) i=-1;
	return(i);
}

void vec_fnerr(double *d, int n, double *a, double **data, int np)
{	// returns a vector of difference between p(x)-y
	int i;
	for(i=0;i<n;i++)d[i]=fnf(a, data[i][0],np)- data[i][1];
}

int fmaxabs(double *d, int n)
{	// returns index of largest element 
	int i, t=0;
	double rmax_a=0.0;
	for(i=0;i<n;i++){
		 if (fabs(d[i])>rmax_a)rmax_a=fabs(d[t=i]);
	}
	return (t);
}

int push_result_to_stack(int n, double abserror, double *pol)
{
	sat_push_real(abserror);
	sat_push_zint_llong(n-1);
    push_vector(n, pol);
	return(0);
}

void							/* initialize constant vector of double */
const_vector(int n, double r, double *v)
{
	int i;
	for (i = 0; i < n; i++)
		v[i] = r;
}




void							// initialize constant matrix of double //
const_matrix(int m, int n, double r, double **mat)
{
	int i,j;
	for (i = 0; i < m; i++){for (j=0;j<n;j++){ mat[i][j]=r;}}
}


int *
create_int_vector(int m)
{
     // Dynamically allocate an n integer array
	int *list,i;
	if(!(list = (int *) calloc(m,sizeof(int))))
	 { error=Insufficient_Memory;return(list);} 
	for (i=0;i<m;i++) {list[i]=0;}
	return (list);
}



double *
create_vector(int m)
{
     // Dynamically allocate an n vector
	double r=0.0;
	double *vec;
	if(!(vec = (double *) calloc(m,sizeof(double)))){  error=Insufficient_Memory;return(vec);} 
	const_vector(m,r,vec); //	for (i=0;i<m;i++) {vec[i]=r;}
	return (vec);
}

double **
create_matrix2(int m, int n)
{
     // Dynamically allocate an mxn 
     // 

	int i;
	const double r=0.0;
	double  **a;
    if(!(a = (double **)calloc(m,sizeof(double *)))) { error=Insufficient_Memory;return(a);} 
    for (i=0;i<m;i++) {
		if (!(a[i] = (double *)calloc(n,sizeof(double)))){a[0]=0; error=Insufficient_Memory;return(a);}
	}
	const_matrix(m,n,r,a); //	for (i=0;i<m;i++){for(j=0;j<n;j++){a[i][j]=r;}}
	return a;
}
void
free2(double *a)    // 
{	
	if(a)free(a);
	a=0;
}
void
free_matrix2(int m,int n,double **a)    // actually only a list of separate vectors ***
{	// free matrix generated by create_matrix2
	int i;
	n=0;   // avoid warning
	if(a){
		for (i=0;i<m;i++){free(a[i]);}
		free(a);
	}
	a=0;
}
void /* backsubstitution */
dgesl(int nb,int n, double **a /* [][n] */ ,int ipvt[],double **b /* [][nb] */,int job)
{
	double t;
	int i,j,k,kb,l,nm1;
	
	nm1 = n - 1;
    if (job == 0) {

/* job = 0, solve a * x = b.
 * First solve l * y = b.
 */
        for (j=0;j<nb;j++){
			if (n > 0) {
				for (k = 0; k < nm1; k++) {
					l = ipvt[k];
					t = b[l][j];
					if (l != k) {
						b[l][j] = b[k][j];
						b[k][j] = t;
					}
/*		saxpy(n-k,t,a(k+1,k),1,b(k+1),1); */
					for (i=k+1;i < n;i++)
						b[i][j] += (t * a[i][k]); 
				}
			}	
         
/* Now solve u * x = y. */
		
			for (kb = 0; kb < n; kb++) {
				k = n - kb-1;
				b[k][j] /= a[k][k];
				t = -b[k][j];
/*		saxpy(k-1,t,a(1,k),1,b(1),1); */
				for (i = 0; i < k ; i++)
					b[i][j] += (t * a[i][k]);
			}
		}
        return;
    }
/* job != 0, solve trans(a) * x = b.
 * First solve trans(u) * x = y.
 */
    for (j=0;j<nb;j++){
		for (k = 0; k < n; k++) {
/* 			t = ddot(k-1,a(1,k),1,b(1),1); */
			t = 0;
			for (i = 0; i < k; i++)
				t += (a[i][k] * b[i][j]);
			b[k][j] = (b[k][j] - t)/a[k][k];
		}
/* Now solve trans(l) * x = y. */
		if (n > 0) {
			for (kb = 0; kb < nm1; kb++) {
				k = n - 2 - kb;
/*			b[k] = b[k] + ddot(n-k,a(k+1,k),1,b(k+1),1); */
				t = 0;
				for (i = k+1;i < n; i++)
					t += (a[i][k] * b[i][j]);
				b[k][j] += t;
				l = ipvt[k];
				if (l != k) {
					t = b[l][j];
					b[l][j] = b[k][j];
					b[k][j] = t;
				}
			}	
        } 
    }
}

/* dgefa.c -- double precision solution of general matrix by
 *	      Gaussian elimination.
 *
 *	This routine performs the LU factorization. (You should
 *	use dgeco.c if you need a condition estimate.)
 *	
 *	Follow this with a call to dgesl.c, which is the linear
 *	equation solver, or dgedi.c which computes the determinant
 *	and inverse.
 *
 *	From "Linpack Users' Guide", Dongarra, et al.
 *	Translated from FORTRAN by C. Bond, 1994.	
 */
// #include <stdio.h>
// #include <math.h>

// #define max(a,b)	(((a) > (b)) ? (a) : (b))
// #define min(a,b)	(((a) < (b)) ? (a) : (b))
 
int /* LU decomposition with pivot */
dgefa(int n, double **a /* [][n] */, int ipvt[])
{
	double dmax,t;
	int i,j,k,kp1,l,nm1, info=0;

	nm1 = n - 1;
    if (n > 0) {
        for (k = 0; k < nm1; k++) {
            kp1 = k + 1;
/* Find l = pivot index. */
            dmax = fabs(a[k][k]);
            l = k;
            for (i = k+1; i < n; i++) {
                if (fabs(a[i][k]) <= dmax) continue;
                l = i;
            }
            ipvt[k] = l;
	
/* Zero pivot implies this column already triangularized. */
            if (a[l][k] == 0.0) {
                info = k;
                continue;
            }

/* Interchange if necessary. */
            if (l != k) {
                t = a[l][k];
                a[l][k] = a[k][k];
                a[k][k] = t;
            }
		
/* Compute multipliers. */
            if ((fabs(a[k][k])) <= 1e-30) { info=-8; goto exit_all;} // printf("\n!ERROR. Singular matrix.\n");
            t = -1.0/a[k][k];
            for (i = k+1; i < n; i++)
                a[i][k] *= t;
		
/* Row elimination with column indexing. */
            for (j = kp1; j < n; j++) {
                t = a[l][j];
                if (l != k) {
                    a[l][j] = a[k][j];
                    a[k][j] = t;
                }
                for (i = k+1; i < n; i++ )
                    a[i][j] += (t * a[i][k]);
            }
        }
    } 
	ipvt[n-1] = n-1;
	if (a[n-1][n-1] == 0.0) info = n-1;
exit_all:
	return info;
}
// ------------------------------------------------------------------------------------------
int								/* pop reals from hp stack into matrix, NB without indices */
pop_matrix_reals(int m, int n, double **mat)
{
	int i, j;
	for (i = m - 1; i >= 0; i--)
	{
		for (j = n - 1; j >= 0; j--)
			mat[i][j] = sat_pop_real();
	}
	return (m * n);
}


int							/* push vector of double */
push_vector(int n, double vec[])
{
	BOOL memOk=0;
	int i,retval=0;
	for (i=0;i<n;i++)	{
		memOk=sat_push_real(vec[i]);
		if (!(memOk)) {error=Insufficient_Memory;retval=DO_ERROR_EXIT; goto exit_all;}
	}
	sat_push_zint_llong(n);
exit_all:
	return retval;
}
